/*
 * linux/sound/arm/ep93xx-alsa-ac97.c -- ALSA PCM interface for the edb93xx ac97 audio
 *
 * Author:      Shrek Wu
 * Created:     July 19, 2005
 * Copyright:   Cirrus Logic, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>

#include <linux/sched.h>
#include <linux/soundcard.h>
#include <linux/version.h>

#include <sound/driver.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/ac97_codec.h>
#include <sound/initval.h>

#include <asm/irq.h>
#include <asm/semaphore.h>
#include <asm/hardware.h>
#include <asm/io.h>
//#include <asm/arch/dma-ep93xx.h>
#include <asm/arch/dma.h>
#include "ep93xx-alsa-pcm.h"


/*------------------------------------------------------------------------------------*/
#ifdef alsa_ac97_debug
#define DEBUG( fmt, arg... )  printk( fmt, ##arg )
#else
#define DEBUG( fmt, arg... )
#endif

/*-------------------------------------------------------------------------------------*/
#define	DRIVER_VERSION	"14-6-2005"
#define	DRIVER_DESC	"EP93xx AC97 Audio driver"

#define AUDIO_NAME      "ep93xx_ac97"
#define AUDIO_SAMPLE_RATE_DEFAULT   44100
static int ac_link_enabled = 0;
static int codec_supported_mixers;


static DECLARE_MUTEX(car_mutex);
static DECLARE_WAIT_QUEUE_HEAD(gsr_wq);
/*static volatile long gsr_bits;*/

static snd_pcm_t *ep93xx_ac97_pcm;
static ac97_t *ep93xx_ac97_ac97;

/*-----------------------------------------------------------------------------------------*/

static int		peek(unsigned int uiAddress);
static int              	poke(unsigned int uiAddress, unsigned int uiValue);
static void     	ep93xx_setup_src(void);
static void     	ep93xx_set_samplerate(long lFrequency, int bCapture);
static void     	ep93xx_init_ac97_codec( void );
static void     	ep93xx_set_hw_format( long format );
static void             ep93xx_init_ac97_controller( void );

static void             ep93xx_audio_enable( int input_or_output_stream );
static void             ep93xx_audio_disable( int input_or_output_stream );

static int 		ep93xx_ac97_pcm_startup(snd_pcm_substream_t *substream);
static void 	ep93xx_ac97_pcm_shutdown(snd_pcm_substream_t *substream);
static int 		ep93xx_ac97_pcm_prepare(snd_pcm_substream_t *substream);
/*----------------------------------------------------------------------------------------*/
static audio_stream_t output_stream =
{
	.audio_num_channels 	=	EP93XX_DEFAULT_NUM_CHANNELS,
	.audio_format 		=	EP93XX_DEFAULT_FORMAT,
	.audio_stream_bitwidth 	=	EP93XX_DEFAULT_BIT_WIDTH,
};

static audio_stream_t input_stream =
{
	.audio_num_channels 	= EP93XX_DEFAULT_NUM_CHANNELS,
	.audio_format 		= EP93XX_DEFAULT_FORMAT,
	.audio_stream_bitwidth 	= EP93XX_DEFAULT_BIT_WIDTH,
};

static audio_state_t audio_state =
{
	.output_stream      	=&output_stream,
	.output_dma        	=DMATx_AAC1,
	.output_id          	="Ac97 out",
	
	.input_stream       	=&input_stream,
	.input_dma          	=DMARx_AAC1,
	.input_id           	="Ac97 in",
	
	.hw_enable          	=ep93xx_audio_enable,
	.hw_disable         	=ep93xx_audio_disable,
	//.client_ioctl       =ep93xx_audio_ioctl,
	
	.set_hw_serial_format	=ep93xx_set_hw_format,
	
	.startup		 = ep93xx_ac97_pcm_startup,
	.shutdown	 = ep93xx_ac97_pcm_shutdown,
	.prepare		 = ep93xx_ac97_pcm_prepare,
	.sem                	 =__MUTEX_INITIALIZER(audio_state.sem),
	
	.DAC_bit_width		 =16,
	.bCompactMode		 =0,
	
};



/*----------------------------------------------------------------------------------------*/
/*
 * peek
 *
 * Reads an AC97 codec register.  Returns -1 if there was an error.
 */
static int peek(unsigned int uiAddress)
{
	unsigned int uiAC97RGIS;
	
	if( !ac_link_enabled )
	{
		printk("ep93xx ac97 peek: attempt to peek before enabling ac-link.\n");
		return -1;
	}
	
	/*
	 * Check to make sure that the address is aligned on a word boundary 
	 * and is 7E or less.
	 */
	if( ((uiAddress & 0x1)!=0) || (uiAddress > 0x007e))
	{
		return -1;
	}

	/*
	 * How it is supposed to work is:
	 *  - The ac97 controller sends out a read addr in slot 1.
	 *  - In the next frame, the codec will echo that address back in slot 1
	 *    and send the data in slot 2.  SLOT2RXVALID will be set to 1.
	 *
	 * Read until SLOT2RXVALID goes to 1.  Reading the data in AC97S2DATA
	 * clears SLOT2RXVALID.
	 */

	/*
	 * First, delay one frame in case of back to back peeks/pokes.
	 */
	mdelay( 1 );

	/*
	 * Write the address to AC97S1DATA, delay 1 frame, read the flags.
	 */
	outl( uiAddress, AC97S1DATA);
	udelay( 21 * 4 );
	uiAC97RGIS = inl( AC97RGIS );

	/*
	 * Return error if we timed out.
	 */
	if( ((uiAC97RGIS & AC97RGIS_SLOT1TXCOMPLETE) == 0 ) &&
		((uiAC97RGIS & AC97RGIS_SLOT2RXVALID) == 0 ) )
	{
		printk( "ep93xx-ac97 - peek failed reading reg 0x%02x.\n", uiAddress ); 
		return -1;
	}
	
	return ( inl(AC97S2DATA) & 0x000fffff);
}

/*
 * poke
 *
 * Writes an AC97 codec Register.  Return -1 if error.
 */
static int poke(unsigned int uiAddress, unsigned int uiValue)
{
	unsigned int uiAC97RGIS;

	if( !ac_link_enabled )
	{
		printk("ep93xx ac97 poke: attempt to poke before enabling ac-link.\n");
		return -1;
	}
	
	/*
	 * Check to make sure that the address is align on a word boundary and
	 * is 7E or less.  And that the value is a 16 bit value.
	 */
	if( ((uiAddress & 0x1)!=0) || (uiAddress > 0x007e))
	{
		printk("ep93xx ac97 poke: address error.\n");
		return -1;
	}
	
	/*
	 * First, delay one frame in case of back to back peeks/pokes.
	 */
	mdelay( 1 );

	/*
	 * Write the data to AC97S2DATA, then the address to AC97S1DATA.
	 */
	outl( uiValue, AC97S2DATA );
	outl( uiAddress, AC97S1DATA );
	
	/*
	 * Wait for the tx to complete, get status.
	 */
	udelay( 30 );/*21*/
	uiAC97RGIS = inl(AC97RGIS);

	/*
	 * Return error if we timed out.
	 */
	if( !(inl(AC97RGIS) & AC97RGIS_SLOT1TXCOMPLETE) )
	{
		printk( "ep93xx-ac97: poke failed writing reg 0x%02x  value 0x%02x.\n", uiAddress, uiValue ); 
		return -1;
	}

	return 0;
}

/*-----------------------------------------------------------------------------------------*/

/*
 * When we get to the multichannel case the pre-fill and enable code
 * will go to the dma driver's start routine.
 */
static void ep93xx_audio_enable( int input_or_output_stream )
{
	unsigned int uiTemp;

	DEBUG("ep93xx_audio_enable :%x\n",input_or_output_stream);

	/*
	 * Enable the rx or tx channel depending on the value of 
	 * input_or_output_stream
	 */
	if( input_or_output_stream )
	{
		uiTemp = inl(AC97TXCR1);
		outl( (uiTemp | AC97TXCR_TEN), AC97TXCR1 );
	}
	else
	{
		uiTemp = inl(AC97RXCR1);
		outl( (uiTemp | AC97RXCR_REN), AC97RXCR1 );
	}
	
	
	//DDEBUG("ep93xx_audio_enable - EXIT\n");
}

static void ep93xx_audio_disable( int input_or_output_stream )
{
	unsigned int uiTemp;

	DEBUG("ep93xx_audio_disable\n");

	/*
	 * Disable the rx or tx channel depending on the value of 
	 * input_or_output_stream
	 */
	if( input_or_output_stream )
	{
		uiTemp = inl(AC97TXCR1);
		outl( (uiTemp & ~AC97TXCR_TEN), AC97TXCR1 );
	}
	else
	{
		uiTemp = inl(AC97RXCR1);
		outl( (uiTemp & ~AC97RXCR_REN), AC97RXCR1 );
	}

	//DDEBUG("ep93xx_audio_disable - EXIT\n");
}


/*
 *cirrus EPD use the cs4202 to be ac97 extern codec
 *The AC97 input use slot 3\4\6\7\8\  to PCM,11    to SPDIF
 *The AC97 input use slot 3\4\6\7\8\9 to PCM,11\10 to SPDIF
 */
/*
static void
ep93xx_set_ac97_slots(unsigned long slots,unsigned short dir)
{
	if(dir==SNDRV_PCM_STREAM_PLAYBACK)	
		outl( slots, AC97TXCR1 );
	else if(dir==SNDRV_PCM_STREAM_CAPTURE)
		outl( slots, AC97RXCR1 );
	else
		printk("%s substream->stream %x error,slots=%x \n",__FUNCTION__,dir,slots);
}

static void
ep93xx_set_ac97_channel(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;

	if(runtime->channels==2)
		ep93xx_set_ac97_slots(AC97_SLOT_4|AC97_SLOT_3,substream->stream);
	else if(runtime->channels==4)
		ep93xx_set_ac97_slots(AC97_SLOT_7|AC97_SLOT_6|AC97_SLOT_4|AC97_SLOT_3,substream->stream);
	else if(runtime->channels==5)
		ep93xx_set_ac97_slots(AC97_SLOT_8|AC97_SLOT_7|AC97_SLOT_6|AC97_SLOT_4|AC97_SLOT_3,substream->stream);
	else if(runtime->channels==6)
		ep93xx_set_ac97_slots(AC97_SLOT_9|AC97_SLOT_8|AC97_SLOT_7|AC97_SLOT_6|AC97_SLOT_4|AC97_SLOT_3,substream->stream);
	else
		printk("%s:ac97 the number of the channel is %x\n",__FUNCTION__,runtime->channels);
}

static void
ep93xx_set_ac97_spdif_channel(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	if(substream->stream==SNDRV_PCM_STREAM_PLAYBACK)
		ep93xx_set_ac97_slots(AC97_SLOT_11|AC97_SLOT_10,substream->stream);
	else if(substream->stream==SNDRV_PCM_STREAM_PLAYBACK)
		ep93xx_set_ac97_slots(AC97_SLOT_11,substream->stream);
}

*/

/*
 * ep93xx_setup_src
 *
 * Once the ac-link is up and all is good, we want to set the codec to a 
 * usable mode.
 */
static void ep93xx_setup_src(void)
{
	int iTemp;

	/*
	 * Set the VRA bit to enable the SRC.
	 */
	iTemp = peek( AC97_2A_EXT_AUDIO_POWER );
	poke( AC97_2A_EXT_AUDIO_POWER,  (iTemp | 0x1) );
	
	/*
	 * Set the DSRC/ASRC bits to enable the variable rate SRC.
	 */
	iTemp = peek( AC97_60_MISC_CRYSTAL_CONTROL  );
	poke( AC97_60_MISC_CRYSTAL_CONTROL, (iTemp  | 0x0300) );
}

/*
 * ep93xx_set_samplerate
 *
 *   lFrequency       - Sample Rate in Hz
 *   bCapture       - 0 to set Tx sample rate; 1 to set Rx sample rate
 */
static void ep93xx_set_samplerate( long lSampleRate, int bCapture )
{
	unsigned short usDivider, usPhase;

	DEBUG( "ep93xx_set_samplerate - Fs = %d\n", (int)lSampleRate );

	if( (lSampleRate <  7200) || (lSampleRate > 48000)  )
	{
		printk( "ep93xx_set_samplerate - invalid Fs = %d\n", 
				 (int)lSampleRate );
		return;
	}

	/*
	 * Calculate divider and phase increment.
	 *
	 * divider = round( 0x1770000 / lSampleRate )
	 *  Note that usually rounding is done by adding 0.5 to a floating 
	 *  value and then truncating.  To do this without using floating
	 *  point, I multiply the fraction by two, do the division, then add one, 
	 *  then divide the whole by 2 and then truncate.
	 *  Same effect, no floating point math.
	 *
	 * Ph incr = trunc( (0x1000000 / usDivider) + 1 )
	 */

	usDivider = (unsigned short)( ((2 * 0x1770000 / lSampleRate) +  1) / 2 );

	usPhase = (0x1000000 / usDivider) + 1;

	/*
	 * Write them in the registers.  Spec says divider must be
	 * written after phase incr.
	 */
	if(!bCapture)
	{
		poke( AC97_2C_PCM_FRONT_DAC_RATE, usDivider);
		poke( AC97_64_DAC_SRC_PHASE_INCR, usPhase);
	}
	else
	{
		
		poke( AC97_32_PCM_LR_ADC_RATE,  usDivider);
		poke( AC97_66_ADC_SRC_PHASE_INCR, usPhase);
	}
	
	DEBUG( "ep93xx_set_samplerate - phase = %d,  divider = %d\n",
				(unsigned int)usPhase, (unsigned int)usDivider );

	/*
	 * We sorta should report the actual samplerate back to the calling
	 * application.  But some applications freak out if they don't get
	 * exactly what they asked for.  So we fudge and tell them what
	 * they want to hear.
	 */
	//audio_samplerate = lSampleRate;

	DEBUG( "ep93xx_set_samplerate - EXIT\n" );
}


/*
 * ep93xx_set_hw_format
 *
 * Sets up whether the controller is expecting 20 bit data in 32 bit words
 * or 16 bit data compacted to have a stereo sample in each 32 bit word.
 */
static void ep93xx_set_hw_format( long format )
{
	int bCompactMode;
	
	switch( format )
	{
		/*
		 * Here's all the <=16 bit formats.  We can squeeze both L and R
		 * into one 32 bit sample so use compact mode.
		 */
		case /*AFMT_U8*/ 	SNDRV_PCM_FMTBIT_U8:		   
		case /*AFMT_S8*/ 	SNDRV_PCM_FMTBIT_S8:		   
		case /*AFMT_S16_LE*/ 	SNDRV_PCM_FMTBIT_S16_LE:
		case /*AFMT_U16_LE*/ 	SNDRV_PCM_FMTBIT_U16_LE:
			bCompactMode = 1;
			break;

		/*
		 * Add any other >16 bit formats here...
		 */
		case /*AFMT_S32_BLOCKED*/ SNDRV_PCM_FMTBIT_S32_LE:
		default:
			bCompactMode = 0;
			break;
	}
	
	if( bCompactMode )
	{
		DEBUG("ep93xx_set_hw_format - Setting serial mode to 16 bit compact.\n");
	
		/*
		 * Turn on Compact Mode so we can fit each stereo sample into
		 * a 32 bit word.  Twice as efficent for DMA and FIFOs.
		 */
		outl( 0x00008018, AC97RXCR1 );
		outl( 0x00008018, AC97TXCR1 );

		audio_state.DAC_bit_width = 16;
		audio_state.bCompactMode = 1;
	}
	else
	{
		DEBUG("ep93xx_set_hw_format - Setting serial mode to 20 bit non-CM.\n");
	
		/*
		 * Turn off Compact Mode so we can do > 16 bits per channel 
		 */
		outl( 0x00004018, AC97RXCR1 );
		outl( 0x00004018, AC97TXCR1 );
		
		audio_state.DAC_bit_width = 20;
		audio_state.bCompactMode = 0;
	}

}


static int
ep93xx_set_ac97_rate(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	int ret=0,reg;
	
	if(runtime->channels==2){
		reg = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
		  	 AC97_PCM_FRONT_DAC_RATE : AC97_PCM_LR_ADC_RATE;
	//	ret =snd_ac97_set_rate(ep93xx_ac97_ac97, reg, runtime->rate);
	ep93xx_set_samplerate(runtime->rate,0);
	ep93xx_set_samplerate(runtime->rate,1);
	}
	else{
		DEBUG("%s:ac97 the number of the channel is %x\n",__FUNCTION__,runtime->channels);
		ret=0;
	}

	return ret;
}



/*
 * ep93xx_init_ac97_controller
 *
 * This routine sets up the Ac'97 Controller.
 */
static void ep93xx_init_ac97_controller(void)
{
	unsigned int uiDEVCFG, uiTemp;

	DEBUG("ep93xx_init_ac97_controller - enter\n");

	/*
	 * Configure the multiplexed Ac'97 pins to be Ac97 not I2s.
	 * Configure the EGPIO4 and EGPIO6 to be GPIOS, not to be  
	 * SDOUT's for the second and third I2S controller channels.
	 */
	uiDEVCFG = inl(SYSCON_DEVCFG);
	
	uiDEVCFG &= ~(SYSCON_DEVCFG_I2SonAC97 | 
				  SYSCON_DEVCFG_A1onG |
				  SYSCON_DEVCFG_A2onG);
		
	SysconSetLocked(SYSCON_DEVCFG, uiDEVCFG);

	/*
	 * Disable the AC97 controller internal loopback.  
	 * Disable Override codec ready.
	 */
	outl( 0, AC97GCR );

	/*
	 * Enable the AC97 Link.
	 */
	uiTemp = inl(AC97GCR);
	outl( (uiTemp | AC97GSR_IFE), AC97GCR );

	/*
	 * Set the TIMEDRESET bit.  Will cause a > 1uSec reset of the ac-link.
	 * This bit is self resetting.
	 */
	outl( AC97RESET_TIMEDRESET, AC97RESET );
	
	/*
	 *  Delay briefly, but let's not hog the processor.
	 */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout( 5 ); /* 50 mSec */

	/*
	 * Read the AC97 status register to see if we've seen a CODECREADY
	 * signal from the AC97 codec.
	 */
	if( !(inl(AC97RGIS) & AC97RGIS_CODECREADY))
	{
		printk( "ep93xx-ac97 - FAIL: CODECREADY still low!\n");
		return;
	}

	/*
	 *  Delay for a second, not hogging the processor
	 */
	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout( HZ ); /* 1 Sec */
	
	/*
	 * Now the Ac-link is up.  We can read and write codec registers.
	 */
	ac_link_enabled = 1;

	/*
	 * Set up the rx and tx channels
	 * Set the CM bit, data size=16 bits, enable tx slots 3 & 4.
	 */
	ep93xx_set_hw_format( EP93XX_DEFAULT_FORMAT );

	DEBUG( "ep93xx-ac97 -- AC97RXCR1:  %08x\n", inl(AC97RXCR1) ); 
	DEBUG( "ep93xx-ac97 -- AC97TXCR1:  %08x\n", inl(AC97TXCR1) ); 

	DEBUG("ep93xx_init_ac97_controller - EXIT - success\n");

}




#define supported_mixer(FOO) \
        ( (FOO >= 0) && \
        (FOO < SOUND_MIXER_NRDEVICES) && \
        codec_supported_mixers & (1<<FOO) )
                                                                                                                             
/*
 * Available record sources.
 * LINE1 refers to AUX in.
 * IGAIN refers to input gain which means stereo mix.
 */
#define AC97_RECORD_MASK \
        (SOUND_MASK_MIC | SOUND_MASK_CD | SOUND_MASK_IGAIN | SOUND_MASK_VIDEO |\
        SOUND_MASK_LINE1 | SOUND_MASK_LINE | SOUND_MASK_PHONEIN)
                                                                                                                             
#define AC97_STEREO_MASK \
        (SOUND_MASK_VOLUME | SOUND_MASK_PCM | SOUND_MASK_LINE | SOUND_MASK_CD | \
        SOUND_MASK_ALTPCM | SOUND_MASK_IGAIN | SOUND_MASK_LINE1 | SOUND_MASK_VIDEO)
                                                                                                                             
#define AC97_SUPPORTED_MASK \
        (AC97_STEREO_MASK | SOUND_MASK_BASS | SOUND_MASK_TREBLE | \
        SOUND_MASK_SPEAKER | SOUND_MASK_MIC | \
        SOUND_MASK_PHONEIN | SOUND_MASK_PHONEOUT)




/* this table has default mixer values for all OSS mixers. */
typedef struct  {
	int mixer;
	unsigned int value;
} mixer_defaults_t;

/*
 * Default mixer settings that are set up during boot.
 *
 * These values are 16 bit numbers in which the upper byte is right volume
 * and the lower byte is left volume or mono volume for mono controls.
 *
 * OSS Range for each of left and right volumes is 0 to 100 (0x00 to 0x64).
 * 
 */
static mixer_defaults_t mixer_defaults[SOUND_MIXER_NRDEVICES] = 
{
	/* Outputs */
	{SOUND_MIXER_VOLUME,	0x6464},   /* 0 dB */  /* -46.5dB to  0 dB */
	{SOUND_MIXER_ALTPCM,	0x6464},   /* 0 dB */  /* -46.5dB to  0 dB */
	{SOUND_MIXER_PHONEOUT,	0x6464},   /* 0 dB */  /* -46.5dB to  0 dB */

	/* PCM playback gain */
	{SOUND_MIXER_PCM,		0x4b4b},   /* 0 dB */  /* -34.5dB to +12dB */

	/* Record gain */
	{SOUND_MIXER_IGAIN,		0x0000},   /* 0 dB */  /* -34.5dB to +12dB */

	/* Inputs */
	{SOUND_MIXER_MIC,		0x0000},   /* mute */  /* -34.5dB to +12dB */
	{SOUND_MIXER_LINE,		0x4b4b},   /* 0 dB */  /* -34.5dB to +12dB */

	/* Inputs that are not connected. */
	{SOUND_MIXER_SPEAKER,	0x0000},   /* mute */  /* -45dB   to   0dB */
	{SOUND_MIXER_PHONEIN,	0x0000},   /* mute */  /* -34.5dB to +12dB */
	{SOUND_MIXER_CD,		0x0000},   /* mute */  /* -34.5dB to +12dB */
	{SOUND_MIXER_VIDEO,		0x0000},   /* mute */  /* -34.5dB to +12dB */
	{SOUND_MIXER_LINE1,		0x0000},   /* mute */  /* -34.5dB to +12dB */

	{-1,0} /* last entry */
};

/* table to scale scale from OSS mixer value to AC97 mixer register value */	
typedef struct {
	unsigned int offset;
	int scale;
} ac97_mixer_hw_t; 

static ac97_mixer_hw_t ac97_hw[SOUND_MIXER_NRDEVICES] = 
{
	[SOUND_MIXER_VOLUME]		=  	{AC97_02_MASTER_VOL,	64},
	[SOUND_MIXER_BASS]			=	{0, 0},
	[SOUND_MIXER_TREBLE]		=	{0, 0},
	[SOUND_MIXER_SYNTH]			=  	{0,	0},
	[SOUND_MIXER_PCM]			=  	{AC97_18_PCM_OUT_VOL,	32},
	[SOUND_MIXER_SPEAKER]		=  	{AC97_0A_PC_BEEP_VOL,	32},
	[SOUND_MIXER_LINE]			=  	{AC97_10_LINE_IN_VOL,	32},
	[SOUND_MIXER_MIC]			=  	{AC97_0E_MIC_VOL,		32},
	[SOUND_MIXER_CD]			=  	{AC97_12_CD_VOL,		32},
	[SOUND_MIXER_IMIX]			=  	{0,	0},
	[SOUND_MIXER_ALTPCM]		=  	{AC97_04_HEADPHONE_VOL,	64},
	[SOUND_MIXER_RECLEV]		=  	{0,	0},
	[SOUND_MIXER_IGAIN]			=  	{AC97_1C_RECORD_GAIN,	16},
	[SOUND_MIXER_OGAIN]			=  	{0,	0},
	[SOUND_MIXER_LINE1]			=  	{AC97_16_AUX_VOL,		32},
	[SOUND_MIXER_LINE2]			=  	{0,	0},
	[SOUND_MIXER_LINE3]			=  	{0,	0},
	[SOUND_MIXER_DIGITAL1]		=  	{0,	0},
	[SOUND_MIXER_DIGITAL2]		=  	{0,	0},
	[SOUND_MIXER_DIGITAL3]		=  	{0,	0},
	[SOUND_MIXER_PHONEIN]		=  	{AC97_0C_PHONE_VOL,		32},
	[SOUND_MIXER_PHONEOUT]		=  	{AC97_06_MONO_VOL,		64},
	[SOUND_MIXER_VIDEO]			=  	{AC97_14_VIDEO_VOL,		32},
	[SOUND_MIXER_RADIO]			=  	{0,	0},
	[SOUND_MIXER_MONITOR]		=  	{0,	0},
};


/* the following tables allow us to go from OSS <-> ac97 quickly. */
enum ac97_recsettings 
{
	AC97_REC_MIC=0,
	AC97_REC_CD,
	AC97_REC_VIDEO,
	AC97_REC_AUX,
	AC97_REC_LINE,
	AC97_REC_STEREO, /* combination of all enabled outputs..  */
	AC97_REC_MONO,	      /*.. or the mono equivalent */
	AC97_REC_PHONE
};

static const unsigned int ac97_rm2oss[] = 
{
	[AC97_REC_MIC] 	 = SOUND_MIXER_MIC,
	[AC97_REC_CD] 	 = SOUND_MIXER_CD,
	[AC97_REC_VIDEO] = SOUND_MIXER_VIDEO,
	[AC97_REC_AUX] 	 = SOUND_MIXER_LINE1,
	[AC97_REC_LINE]  = SOUND_MIXER_LINE,
	[AC97_REC_STEREO]= SOUND_MIXER_IGAIN,
	[AC97_REC_PHONE] = SOUND_MIXER_PHONEIN
};

/* indexed by bit position */
static const unsigned int ac97_oss_rm[] = 
{
	[SOUND_MIXER_MIC] 	= AC97_REC_MIC,
	[SOUND_MIXER_CD] 	= AC97_REC_CD,
	[SOUND_MIXER_VIDEO] = AC97_REC_VIDEO,
	[SOUND_MIXER_LINE1] = AC97_REC_AUX,
	[SOUND_MIXER_LINE] 	= AC97_REC_LINE,
	[SOUND_MIXER_IGAIN]	= AC97_REC_STEREO,
	[SOUND_MIXER_PHONEIN] 	= AC97_REC_PHONE
};


/* reads the given OSS mixer from the ac97 the caller must have insured that the ac97 knows
   about that given mixer, and should be holding a spinlock for the card */
/*
static int ep93xx_read_mixer( int oss_channel ) 
{
	u16 val;
	int ret = 0;
	ac97_mixer_hw_t * mh = &ac97_hw[oss_channel];

	if( !mh->scale )
	{
		printk( "ep93xx-ac97.c: ep93xx_read_mixer - not a valid OSS channel\n");
		return 0;
	}

	val = peek( mh->offset );
	
	if( val & 0x8000 ) 
	{
		ret = 0;
	}
	else if (AC97_STEREO_MASK & (1 << oss_channel)) 
	{
		int left,right;

		left = (val >> 8)  & 0x7f;
		right = val  & 0x7f;

		right = 100 - ((right * 100) / mh->scale);
		left = 100 - ((left * 100) / mh->scale);

		ret = ((right << 8) | left);
	}
	else if (oss_channel == SOUND_MIXER_SPEAKER) 
	{
		ret = 100 - ((((val & 0x1e)>>1) * 100) / mh->scale);
	}
	else if( (oss_channel == SOUND_MIXER_PHONEIN) ||
		(oss_channel == SOUND_MIXER_PHONEOUT) ||
		(oss_channel == SOUND_MIXER_MIC) )
	{
		ret = 100 - (((val & 0x1f) * 100) / mh->scale);
	} 
	//  
	// For bass and treble, the low bit is optional.  Masking it
	// lets us avoid the 0xf 'bypass'.. 
	//
	else if (oss_channel == SOUND_MIXER_BASS) 
	{
		ret = 100 - ((((val >> 8) & 0xe) * 100) / mh->scale);
	} 
	else if (oss_channel == SOUND_MIXER_TREBLE) 
	{
		ret = 100 - (((val & 0xe) * 100) / mh->scale);
	}
	
	DEBUG("ac97_codec: read OSS mixer %2d (ac97 register 0x%02x), "
	       "0x%04x -> 0x%04x\n",
	       oss_channel, mh->offset, val, ret);

	return ret;
}
*/

/*
 * ep93xx_write_mixer
 *
 */
static void ep93xx_write_mixer
( 
	int oss_channel,
	unsigned int left, 
	unsigned int right
)
{
	u16 val = 0;
	ac97_mixer_hw_t * mh = &ac97_hw[oss_channel];

	DEBUG("ac97_codec: wrote OSS %2d (ac97 0x%02x), "
	       "l:%2d, r:%2d:",
	       oss_channel, mh->offset, left, right);

	if( !mh->scale )
	{
		printk( "ep93xx-ac97.c: ep93xx_write_mixer - not a valid OSS channel\n");
		return;
	}

	if( AC97_STEREO_MASK & (1 << oss_channel) ) 
	{
		/* stereo mixers */
		if (left == 0 && right == 0) 
		{
			val = 0x8000;
		} 
		else 
		{
			if (oss_channel == SOUND_MIXER_IGAIN) 
			{
				right = (right * mh->scale) / 100;
				left = (left * mh->scale) / 100;
				if (right >= mh->scale)
					right = mh->scale-1;
				if (left >= mh->scale)
					left = mh->scale-1;
			} 
			else 
			{
				right = ((100 - right) * mh->scale) / 100;
				left = ((100 - left) * mh->scale) / 100;
				if (right >= mh->scale)
					right = mh->scale-1;
				if (left >= mh->scale)
					left = mh->scale-1;
			}
			val = (left << 8) | right;
		}
	} 
	else if(left == 0) 
	{
		val = 0x8000;
	} 
	else if( (oss_channel == SOUND_MIXER_SPEAKER) ||
			(oss_channel == SOUND_MIXER_PHONEIN) ||
			(oss_channel == SOUND_MIXER_PHONEOUT) )
	{
		left = ((100 - left) * mh->scale) / 100;
		if (left >= mh->scale)
			left = mh->scale-1;
		val = left;
	} 
	else if (oss_channel == SOUND_MIXER_MIC) 
	{
		val = peek( mh->offset) & ~0x801f;
		left = ((100 - left) * mh->scale) / 100;
		if (left >= mh->scale)
			left = mh->scale-1;
		val |= left;
	} 
	/*  
	 * For bass and treble, the low bit is optional.  Masking it
	 * lets us avoid the 0xf 'bypass'.
	 * Do a read, modify, write as we have two contols in one reg. 
	 */
	else if (oss_channel == SOUND_MIXER_BASS) 
	{
		val = peek( mh->offset) & ~0x0f00;
		left = ((100 - left) * mh->scale) / 100;
		if (left >= mh->scale)
			left = mh->scale-1;
		val |= (left << 8) & 0x0e00;
	} 
	else if (oss_channel == SOUND_MIXER_TREBLE) 
	{
		val = peek( mh->offset) & ~0x000f;
		left = ((100 - left) * mh->scale) / 100;
		if (left >= mh->scale)
			left = mh->scale-1;
		val |= left & 0x000e;
	}
	
	DEBUG(" 0x%04x", val);

	poke( mh->offset, val );

#ifdef alsa_ac97_debug
	val = peek( mh->offset );
	DEBUG(" -> 0x%04x\n", val);
#endif

}

/* a thin wrapper for write_mixer */
static void ep93xx_set_mixer
(
	unsigned int oss_mixer, 
	unsigned int val 
) 
{
	unsigned int left,right;

	/* cleanse input a little */
	right = ((val >> 8)  & 0xff) ;
	left = (val  & 0xff) ;

	if (right > 100) right = 100;
	if (left > 100) left = 100;

	/*mixer_state[oss_mixer] = (right << 8) | left;*/
	ep93xx_write_mixer( oss_mixer, left, right);
}

static void ep93xx_init_mixer(void)
{
	u16 cap;
	int i;

	/* mixer masks */
	codec_supported_mixers 	= AC97_SUPPORTED_MASK;
	
	cap = peek( AC97_00_RESET );
	if( !(cap & 0x04) )
	{
		codec_supported_mixers &= ~(SOUND_MASK_BASS|SOUND_MASK_TREBLE);
	}
	if( !(cap & 0x10) )
	{
		codec_supported_mixers &= ~SOUND_MASK_ALTPCM;
	}

	/* 
	 * Detect bit resolution of output volume controls by writing to the
	 * 6th bit (not unmuting yet)
	 */
	poke( AC97_02_MASTER_VOL, 0xa020 );
	if( peek( AC97_02_MASTER_VOL) != 0xa020 )
	{
		ac97_hw[SOUND_MIXER_VOLUME].scale = 32;
	}

	poke( AC97_04_HEADPHONE_VOL, 0xa020 );
	if( peek( AC97_04_HEADPHONE_VOL) != 0xa020 )
	{
		ac97_hw[AC97_04_HEADPHONE_VOL].scale = 32;
	}

	poke( AC97_06_MONO_VOL, 0x8020 );
	if( peek( AC97_06_MONO_VOL) != 0x8020 )
	{
		ac97_hw[AC97_06_MONO_VOL].scale = 32;
	}

	/* initialize mixer channel volumes */
	for( i = 0; 
		(i < SOUND_MIXER_NRDEVICES) && (mixer_defaults[i].mixer != -1) ; 
		i++ ) 
	{
		if( !supported_mixer( mixer_defaults[i].mixer) )
		{ 
			continue;
		}
		
		ep93xx_set_mixer( mixer_defaults[i].mixer, mixer_defaults[i].value);
	}

}


/*
 * ac97_recmask_io
 *
 * Read or write the record source.
 */
/*
static int ep93xx_read_recsource(void) 
{
	unsigned int val;

	// read it from the card 
	val = peek( AC97_1A_RECORD_SELECT );
	
	DEBUG("ac97_codec: ac97 recmask to set to 0x%04x\n", val);
	
	return( 1 << ac97_rm2oss[val & 0x07] );
}
*/

static int ep93xx_set_recsource( int mask ) 
{
	unsigned int val;

	/* Arg contains a bit for each recording source */
	if( mask == 0 ) 
	{
		return 0;
	}
	
	mask &= AC97_RECORD_MASK;
	
	if( mask == 0 ) 
	{
		return -EINVAL;
	}
				
	/*
	 * May have more than one bit set.  So clear out currently selected
	 * record source value first (AC97 supports only 1 input) 
	 */
	val = (1 << ac97_rm2oss[peek( AC97_1A_RECORD_SELECT ) & 0x07]);
	if (mask != val)
	    mask &= ~val;
       
	val = ffs(mask); 
	val = ac97_oss_rm[val-1];
	val |= val << 8;  /* set both channels */

	DEBUG("ac97_codec: setting ac97 recmask to 0x%04x\n", val);

	poke( AC97_1A_RECORD_SELECT, val);

	return 0;
}






/*
 * ep93xx_init_ac97_codec
 *
 * Program up the external Ac97 codec.
 *
 */
static void ep93xx_init_ac97_codec( void )
{
	DEBUG("ep93xx_init_ac97_codec - enter\n");

	ep93xx_setup_src();
	ep93xx_set_samplerate( AUDIO_SAMPLE_RATE_DEFAULT, 0 );
	ep93xx_set_samplerate( AUDIO_SAMPLE_RATE_DEFAULT, 1 );
	ep93xx_init_mixer();

	DEBUG("ep93xx_init_ac97_codec - EXIT\n");
	
}

#if alsa_ac97_debug
static void ep93xx_dump_ac97_regs(void)
{
	int i;
	unsigned int reg0, reg1, reg2, reg3, reg4, reg5, reg6, reg7;
	
	DEBUG( "---------------------------------------------\n");
	DEBUG( "   :   0    2    4    6    8    A    C    E\n" ); 
	
	for( i=0 ; i < 0x80 ; i+=0x10 )
	{
		reg0 = 0xffff & (unsigned int)peek( i );
		reg1 = 0xffff & (unsigned int)peek( i + 0x2 );
		reg2 = 0xffff & (unsigned int)peek( i + 0x4 );
		reg3 = 0xffff & (unsigned int)peek( i + 0x6 );
		reg4 = 0xffff & (unsigned int)peek( i + 0x8 );
		reg5 = 0xffff & (unsigned int)peek( i + 0xa );
		reg6 = 0xffff & (unsigned int)peek( i + 0xc );
		reg7 = 0xffff & (unsigned int)peek( i + 0xe );

		DEBUG( " %02x : %04x %04x %04x %04x %04x %04x %04x %04x\n", 
				 i, reg0, reg1, reg2, reg3, reg4, reg5, reg6, reg7);
	}
}
#endif

/*
 * Audio interface
 */
static void ep93xx_audio_init(void *dummy)
{
	DEBUG("ep93xx_audio_init - enter\n");
	
	/*
	 * Init the controller, enable the ac-link.
	 * Initialize the codec.
	 */
	ep93xx_init_ac97_controller();
	ep93xx_init_ac97_codec();
#if alsa_ac97_debug
	ep93xx_dump_ac97_regs();
#endif
	DEBUG("ep93xx_audio_init - EXIT\n");
}

/*----------------------------------------------------------------------------------------*/
static unsigned short ep93xx_ac97_read(ac97_t *ac97, unsigned short reg)
{
	int val = -1;
	/*volatile u32 *reg_addr;*/

	DEBUG(" number of codec:%x reg=%x\n",ac97->num,reg);
	val=peek(reg);
	if(val==-1){
		printk(KERN_ERR "%s: read error (ac97_reg=%d )val=%x\n",
				__FUNCTION__, reg, val);
		return 0;
	}

	return val;
}

static void ep93xx_ac97_write(ac97_t *ac97, unsigned short reg, unsigned short val)
{
	/*volatile u32 *reg_addr;*/
	int ret;

	DEBUG(" number of codec:%x rge=%x val=%x\n",ac97->num,reg,val);
	ret=poke(reg, val);
	if(ret!=0){
		printk(KERN_ERR "%s: write error (ac97_reg=%d val=%x)\n",
				__FUNCTION__, reg, val);
	}

}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
static void ep93xx_ac97_reset(ac97_t *ac97)
{

	DEBUG(" ep93xx_ac97_reset\n");
	ep93xx_audio_init(0);

}

static ac97_bus_ops_t ep93xx_ac97_ops = {
	.read	= ep93xx_ac97_read,
	.write	= ep93xx_ac97_write,
	.reset	= ep93xx_ac97_reset,
};
#endif


static int ep93xx_ac97_pcm_startup(snd_pcm_substream_t *substream)
{
	snd_pcm_runtime_t *runtime = substream->runtime;
	int r;

	runtime->hw.channels_min = 2;
	runtime->hw.channels_max = 2;

	r = (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) ?
	    AC97_RATES_FRONT_DAC : AC97_RATES_ADC;
	
	DEBUG(" ep93xx_ac97_pcm_startup=%x\n",r);

		return 0;
}

static void ep93xx_ac97_pcm_shutdown(snd_pcm_substream_t *substream)
{
	DEBUG(" ep93xx_ac97_pcm_shutdown\n");

}

static int ep93xx_ac97_pcm_prepare(snd_pcm_substream_t *substream)
{
	int ret=-1;

	ret=ep93xx_set_ac97_rate(substream);
	if(substream->stream == SNDRV_PCM_STREAM_CAPTURE){
		ep93xx_set_recsource(SOUND_MASK_MIC|SOUND_MASK_LINE1 | SOUND_MASK_LINE);
		poke( AC97_1C_RECORD_GAIN, 0);
	}
#if alsa_ac97_debug
	ep93xx_dump_ac97_regs();
#endif

	DEBUG(" ep93xx_ac97_pcm_prepare=%x\n",ret);
	return ret;
}

/*----------------------------------------------------------------------------------------------*/

#ifdef CONFIG_PM

static int ep93xx_ac97_do_suspend(snd_card_t *card, unsigned int state)
{
	if (card->power_state != SNDRV_CTL_POWER_D3cold) {
		snd_pcm_suspend_all(ep93xx_ac97_pcm);
		snd_ac97_suspend(ep93xx_ac97_ac97);
		snd_power_change_state(card, SNDRV_CTL_POWER_D3cold);
	}

	return 0;
}

static int ep93xx_ac97_do_resume(snd_card_t *card, unsigned int state)
{
	if (card->power_state != SNDRV_CTL_POWER_D0) {

		snd_ac97_resume(ep93xx_ac97_ac97);
		snd_power_change_state(card, SNDRV_CTL_POWER_D0);
	}

	return 0;
}

static int ep93xx_ac97_suspend(struct device *_dev, u32 state, u32 level)
{
	snd_card_t *card = dev_get_drvdata(_dev);
	int ret = 0;

	if (card && level == SUSPEND_DISABLE)
		ret = ep93xx_ac97_do_suspend(card, SNDRV_CTL_POWER_D3cold);

	return ret;
}

static int ep93xx_ac97_resume(struct device *_dev, u32 level)
{
	snd_card_t *card = dev_get_drvdata(_dev);
	int ret = 0;

	if (card && level == RESUME_ENABLE)
		ret = ep93xx_ac97_do_resume(card, SNDRV_CTL_POWER_D0);

	return ret;
}

#else

#define ep93xx_ac97_do_suspend		NULL
#define ep93xx_ac97_do_resume		NULL
#define ep93xx_ac97_suspend		NULL
#define ep93xx_ac97_resume		NULL

#endif

static int ep93xx_ac97_probe(struct device *dev)
{

	snd_card_t *card;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	
	ac97_bus_t *ac97_bus;
	ac97_template_t ac97_template;

#else
	ac97_bus_t ac97_bus_ops, *ac97_bus;
	ac97_t ac97;
#endif
	int ret;

	/*
	 * Enable audio early on, give the DAC time to come up.
	 */ 
	ep93xx_audio_init( 0 );
	

	ret = -ENOMEM;
	/*regist the new card device*/
	card = snd_card_new(SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1,
			    THIS_MODULE, 0);
	if (!card){
		printk("AC97: snd_card_new error\n");
		goto err;
	}

	card->dev = dev;
	
	/*regist the new pcm device*/
	ret = ep93xx_ac97_pcm_new(card, &audio_state, &ep93xx_ac97_pcm);
	if (ret){
		printk("AC97: ep93xx_ac97_pcm_new error\n");
		goto err;
	}
		

	
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	/*regist the new ac97 device*/
	ret = snd_ac97_bus(card, 0, &ep93xx_ac97_ops, NULL, &ac97_bus);
	if (ret){
		printk("AC97: snd_ac97_bus error\n");
		goto err;
	}
	memset(&ac97_template, 0, sizeof(ac97_template));
	ret = snd_ac97_mixer(ac97_bus, &ac97_template, &ep93xx_ac97_ac97);
	if (ret){
		printk("AC97: snd_ac97_mixer error\n");
		goto err;
	}
	
	strncpy(card->driver, dev->driver->name, sizeof(card->driver));

	snprintf(card->shortname, sizeof(card->shortname),
		 "%s", snd_ac97_get_short_name(ep93xx_ac97_ac97));
	snprintf(card->longname, sizeof(card->longname),
		 "%s (%s)", dev->driver->name, card->mixername);
	
#else
	memset(&ac97_bus_ops, 0, sizeof(ac97_bus_ops));
	ac97_bus_ops.write = ep93xx_ac97_write;
	ac97_bus_ops.read = ep93xx_ac97_read;
	
	ret = snd_ac97_bus(card, &ac97_bus_ops, &ac97_bus);
	if (ret){
		printk("AC97: snd_ac97_bus error\n");
		goto err;
	}

	DEBUG("AC97: register snd_ac97_bus ok\n");
	
	memset(&ac97, 0, sizeof(ac97));
	ret = snd_ac97_mixer(ac97_bus, &ac97, &ep93xx_ac97_ac97);
	if (ret){
		printk("AC97: snd_ac97_mixer error\n");
		goto err;
	}
	
	DEBUG("AC97: register snd_ac97_mixer ok\n");

	strcpy(card->driver,    "EP93xx  CS4202");
	strcpy(card->shortname, "EPD93XX CS4202");
	strcpy(card->longname,  "Cirrus EPD93XX Codec CS4202");
	
#endif

#if alsa_ac97_debug
	ep93xx_dump_ac97_regs();
#endif

	ep93xx_audio_init( 0 );	

	/*setting the card device callback*/
        	snd_card_set_pm_callback(card, ep93xx_ac97_do_suspend,
                                 ep93xx_ac97_do_resume, NULL);

	/*regist the new CARD device*/
	ret = snd_card_register(card);
	if (ret == 0) {
		dev_set_drvdata(dev, card);
		printk("AC97: register snd_card_register ok\n");
		return 0;
	}

 err:
	if (card){
		snd_card_free(card);

		DEBUG("the card register fail");
	}

	return ret;
}

static int ep93xx_ac97_remove(struct device *dev)
{
	snd_card_t *card = dev_get_drvdata(dev);

	if (card) {
		snd_card_free(card);
		dev_set_drvdata(dev, NULL);
	}

	return 0;
}

static struct device_driver ep93xx_ac97_driver = {
	.name		= "ep93xx-ac97",
	.bus		= &platform_bus_type,
	.probe		= ep93xx_ac97_probe,
	.remove		= ep93xx_ac97_remove,
	.suspend		= ep93xx_ac97_suspend,
	.resume		= ep93xx_ac97_resume,
};

static int __init ep93xx_ac97_init(void)
{
	DEBUG(KERN_INFO "%s: version %s\n", DRIVER_DESC, DRIVER_VERSION);
	return driver_register(&ep93xx_ac97_driver);
}

static void __exit ep93xx_ac97_exit(void)
{
	driver_unregister(&ep93xx_ac97_driver);
}

module_init(ep93xx_ac97_init);
module_exit(ep93xx_ac97_exit);

MODULE_AUTHOR("Shrek");
MODULE_DESCRIPTION("AC97 driver for the Cirrus EP93xx chip");
MODULE_LICENSE("GPL");
